import os
import re
import time
import logging
from django.conf import settings as django_settings
from django.db.models import F
from django.http import HttpResponseForbidden
from django.utils import timezone

import geoip2.database
import geoip2.errors

from .models import (
    Visitor, PageView, SecurityEvent,
    ThreatScore, IPWhitelist, HoneypotHit
)
from .conf import monitor_settings, is_redis_available,is_celery_available

logger = logging.getLogger('django_security_monitor')

# ── Attack-detection compiled patterns ──────────────────────────────────────
_SQL = re.compile(
    r"(union[\s+]+select|select[\s\S]+from|insert[\s+]+into|drop[\s+]+table|"
    r"or[\s+]+1[\s+]*=[\s+]*1|exec[\s+]*\(|xp_cmdshell|information_schema|"
    r"sleep[\s+]*\(|benchmark[\s+]*\(|load_file[\s+]*\(|extractvalue[\s+]*\()",
    re.IGNORECASE
)
_XSS = re.compile(
    r"(<script[\s>]|javascript\s*:|on\w+\s*=\s*['\"]|<iframe[\s>]|"
    r"eval\s*\(|document\.cookie|window\.location|<svg[\s\S]*?on\w+\s*=)",
    re.IGNORECASE
)
_TRAVERSAL = re.compile(
    r"(\.\./|\.\.\\|%2e%2e%2f|%2e%2e/|\.\.%2f|%252e%252e|\.\.;/)",
    re.IGNORECASE
)
_SCANNER_UA = re.compile(
    r"(nikto|sqlmap|nmap|masscan|zgrab|nuclei|gobuster|dirbuster|"
    r"wfuzz|burpsuite|metasploit|acunetix|nessus|openvas|whatweb|"
    r"hydra|medusa|w3af|skipfish|arachni|zaproxy|havij|commix)",
    re.IGNORECASE
)


class SecurityMonitorMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self._use_redis = is_redis_available()
        self._whitelist: set = set()
        self._whitelist_ts: float = 0.0

    def __call__(self, request):
        path = request.path

        # 1. Skip excluded paths
        if any(path.startswith(p) for p in monitor_settings.EXCLUDE_PATHS):
            return self.get_response(request)

        ip = self._get_ip(request)
        self._refresh_whitelist()

        # 2. Skip whitelisted IPs
        if ip in self._whitelist:
            return self.get_response(request)

        # 3. Superusers bypass the blocked-IP gate
        user = getattr(request, 'user', None)
        is_superuser = bool(user and user.is_authenticated and user.is_superuser)

        if not is_superuser:
            threat = self._get_threat(ip)
            if threat and threat.is_blocked:
                self._log(ip, request, 'blocked_ip', 'critical', 0)
                return HttpResponseForbidden(
                    "⛔ Your access has been suspended due to suspicious activity."
                )

        # 4. Run security analysis
        self._analyze(ip, request)

        # 5. Get response and log HTTP errors
        response = self.get_response(request)

        if response.status_code == 404 and monitor_settings.LOG_404:
            self._log(ip, request, '404', 'low',
                    monitor_settings.THREAT_SCORE_WEIGHTS.get('404', 3))
        elif response.status_code == 500 and monitor_settings.LOG_500:
            self._log(ip, request, '500', 'low',
                    monitor_settings.THREAT_SCORE_WEIGHTS.get('500', 2))

        return response


    # ── analysis ───────────────────────────────────────────────────────────
    def _analyze(self, ip, request):
        path = request.path
        path_lower = path.lower()
        ua = request.META.get('HTTP_USER_AGENT', '')
        qs = request.META.get('QUERY_STRING', '')
        full = path_lower + qs
        w = monitor_settings.THREAT_SCORE_WEIGHTS

        # 1. Honeypot
        if path in monitor_settings.HONEYPOT_PATHS:
            HoneypotHit.objects.create(
                ip_address=ip, path=path,
                user_agent=ua[:500],
                headers={k: v for k, v in request.META.items()
                         if k.startswith('HTTP_')}
            )
            self._log(ip, request, 'honeypot', 'critical', w.get('honeypot', 40))
            return  # no further checks — we already know this is bad

        # 2. Scanner UA
        if _SCANNER_UA.search(ua):
            self._log(ip, request, 'scanner', 'high', w.get('scanner', 20))

        # 3. Suspicious paths
        for sp in monitor_settings.SUSPICIOUS_PATHS:
            if sp.lower() in path_lower:
                self._log(ip, request, 'suspicious_path', 'high',
                          w.get('suspicious_path', 15))
                break

        # 4. Sensitive extensions
        for ext in monitor_settings.SENSITIVE_EXTENSIONS:
            if path_lower.endswith(ext.lower()):
                self._log(ip, request, 'sensitive_file', 'critical',
                          w.get('sensitive_file', 20))
                break

        # 5. Path traversal
        if _TRAVERSAL.search(full):
            self._log(ip, request, 'path_traversal', 'critical',
                      w.get('path_traversal', 35))

        # 6. SQL injection
        if _SQL.search(qs):
            self._log(ip, request, 'sql_injection', 'critical',
                      w.get('sql_injection', 30), payload=qs[:500])

        # 7. XSS
        if _XSS.search(qs):
            self._log(ip, request, 'xss_attempt', 'high',
                      w.get('xss_attempt', 25), payload=qs[:500])

        # 8. Rate limiting
        if self._use_redis:
            self._rate_redis(ip, request)
        else:
            self._rate_db(ip, request)

    # ── rate limiting ───────────────────────────────────────────────────────
    def _rate_redis(self, ip, request):
        if monitor_settings.RATE_LIMIT is None or monitor_settings.RATE_LIMIT_WINDOW is None:
            return
        from django.core.cache import cache
        key = f'sm_rl_{ip}'
        count = cache.get(key, 0)
        if count == 0:
            cache.set(key, 1, monitor_settings.RATE_LIMIT_WINDOW)
        else:
            cache.incr(key)
            if count + 1 > monitor_settings.RATE_LIMIT:
                w = monitor_settings.THREAT_SCORE_WEIGHTS
                self._log(ip, request, 'rate_limit', 'medium', w.get('rate_limit', 10))

    def _rate_db(self, ip, request):
        if monitor_settings.RATE_LIMIT is None or monitor_settings.RATE_LIMIT_WINDOW is None:
            return
        from datetime import timedelta
        window_start = timezone.now() - timedelta(seconds=monitor_settings.RATE_LIMIT_WINDOW)
        count = SecurityEvent.objects.filter(
            ip_address=ip, timestamp__gte=window_start
        ).count()
        if count > monitor_settings.RATE_LIMIT:
            w = monitor_settings.THREAT_SCORE_WEIGHTS
            self._log(ip, request, 'rate_limit', 'medium', w.get('rate_limit', 10))

    # ── helpers ─────────────────────────────────────────────────────────────
    def _log(self, ip, request, event_type, severity, delta, payload=''):
        user = getattr(request, 'user', None)
        if user and not user.is_authenticated:
            user = None

        # Check if superuser BEFORE creating the event
        is_superuser = bool(user and user.is_superuser)

        SecurityEvent.objects.create(
            ip_address=ip,
            user=user,
            event_type=event_type,
            severity=severity,
            path=request.path[:500],
            method=request.method,
            user_agent=request.META.get('HTTP_USER_AGENT', '')[:500],
            referer=request.META.get('HTTP_REFERER', '')[:500],
            payload=payload,
            threat_score_delta=delta,
            # ✅ tag the event so dashboard can show [TEST] label
            extra_data={'superuser_test': True} if is_superuser else {},
        )
        self._bump_score(ip, delta, is_superuser=is_superuser)


    def _bump_score(self, ip, delta, is_superuser=False):
        if delta == 0:
            return
        threat, created = ThreatScore.objects.get_or_create(
            ip_address=ip, defaults={'score': delta, 'event_count': 1}
        )
        if not created:
            ThreatScore.objects.filter(pk=threat.pk).update(
                score=F('score') + delta,
                event_count=F('event_count') + 1,
            )
        threat.refresh_from_db()

        # ✅ Never auto-block a superuser — score is tracked for testing purposes
        if is_superuser:
            return

        if (
            monitor_settings.AUTO_BLOCK
            and monitor_settings.BLOCK_THRESHOLD is not None
            and threat.score >= monitor_settings.BLOCK_THRESHOLD
            and not threat.is_blocked
        ):
            ThreatScore.objects.filter(pk=threat.pk).update(
                is_blocked=True,
                blocked_at=timezone.now(),
                block_reason=f'Auto-blocked: score ≥ {monitor_settings.BLOCK_THRESHOLD}',
            )



    def _get_threat(self, ip):
        try:
            return ThreatScore.objects.get(ip_address=ip)
        except ThreatScore.DoesNotExist:
            return None

    def _refresh_whitelist(self):
        ttl = monitor_settings.WHITELIST_CACHE_TTL
        if time.time() - self._whitelist_ts > ttl:
            try:
                self._whitelist = set(
                    IPWhitelist.objects.values_list('ip_address', flat=True)
                )
            except Exception:
                pass
            self._whitelist_ts = time.time()

    def _get_ip(self, request):
        xff = request.META.get('HTTP_X_FORWARDED_FOR')
        if xff:
            return xff.split(',')[0].strip()
        xri = request.META.get('HTTP_X_REAL_IP')
        if xri:
            return xri.strip()
        return request.META.get('REMOTE_ADDR', '')


class VisitorTrackingMiddleware:
    """Your existing middleware — enhanced with response time & status tracking."""

    def __init__(self, get_response):
        self.get_response = get_response
        self.geoip_reader = None
        geoip_path = getattr(django_settings, 'GEOIP_PATH', None)
        if geoip_path:
            db = os.path.join(geoip_path, 'GeoLite2-City.mmdb')
            try:
                self.geoip_reader = geoip2.database.Reader(db)
            except FileNotFoundError:
                pass

    def __call__(self, request):
        # 1. Skip paths that don't need visitor tracking
        skip = list(monitor_settings.EXCLUDE_PATHS) + [
            '/admin/', '/newadmin/', '/prometheus/'
        ]
        if any(request.path.startswith(p) for p in skip):
            return self.get_response(request)

        # 2. Ensure session exists
        if not request.session.session_key:
            request.session.create()
        session_key = request.session.session_key

        # 3. Get IP and UA
        ip = self._get_ip(request) or ''
        ua = request.META.get('HTTP_USER_AGENT', '')[:500]

        # 4. Get or create visitor record
        visitor, created = Visitor.objects.get_or_create(
            session_key=session_key,
            defaults={'ip_address': ip, 'user_agent': ua}
        )

        # 5. Link to auth user if logged in
        if request.user.is_authenticated and visitor.user_id != request.user.id:
            visitor.user = request.user
            visitor.save(update_fields=['user'])

        # 6. Populate location on first visit, otherwise bump count
        if created:
            self._geo(visitor, ip)
        else:
            Visitor.objects.filter(pk=visitor.pk).update(
                visit_count=F('visit_count') + 1,
                last_visit=timezone.now(),
            )

        # 7. Measure response time
        t0 = time.time()
        response = self.get_response(request)
        elapsed = int((time.time() - t0) * 1000)

        # 8. Log the page view
        if is_celery_available():
            from .tasks import log_pageview_async

            log_pageview_async.delay(
                visitor_id=visitor.pk,
                path=request.path[:500],
                method=request.method,
                status_code=response.status_code,
                response_time_ms=elapsed,
            )
        else:
            PageView.objects.create(
                visitor=visitor,
                path=request.path[:500],
                method=request.method,
                status_code=response.status_code,
                response_time_ms=elapsed,
        )

        return response


    def _geo(self, visitor, ip):
        if not self.geoip_reader or not ip:
            return
        try:
            rec = self.geoip_reader.city(ip)
            visitor.location = {
                'country': rec.country.name,
                'country_code': rec.country.iso_code,
                'city': rec.city.name,
                'postal_code': rec.postal.code,
                'latitude': rec.location.latitude,
                'longitude': rec.location.longitude,
            }
            visitor.save(update_fields=['location'])
        except (geoip2.errors.AddressNotFoundError,
                geoip2.errors.GeoIP2Error, ValueError):
            pass

    def _get_ip(self, request):
        xff = request.META.get('HTTP_X_FORWARDED_FOR')
        if xff:
            return xff.split(',')[0].strip()
        xri = request.META.get('HTTP_X_REAL_IP')
        if xri:
            return xri.strip()
        return request.META.get('REMOTE_ADDR')



from datetime import timedelta
from django.contrib.auth.signals import user_logged_in, user_login_failed
from django.db.models import F
from django.dispatch import receiver
from django.utils import timezone

from .models import LoginAttempt, SecurityEvent, ThreatScore
from .conf import monitor_settings


def _get_ip(request):
    if not request:
        return ''
    xff = request.META.get('HTTP_X_FORWARDED_FOR')
    return xff.split(',')[0].strip() if xff else request.META.get('REMOTE_ADDR', '')


@receiver(user_login_failed)
def on_login_failed(sender, credentials, request, **kwargs):
    if not request:
        return
    ip = _get_ip(request)
    username = credentials.get('username', '')
    ua = request.META.get('HTTP_USER_AGENT', '')[:500]
    w = monitor_settings.THREAT_SCORE_WEIGHTS

    # ✅ Check if the failing username belongs to a superuser
    from django.contrib.auth import get_user_model
    User = get_user_model()
    is_superuser = User.objects.filter(
        username=username, is_superuser=True
    ).exists()

    LoginAttempt.objects.create(
        ip_address=ip, username=username, success=False, user_agent=ua
    )

    if monitor_settings.LOGIN_ATTEMPT_WINDOW is None:
        # If disabled, just log the single failure as an auth_failure
        failures = 0
    else:
        window_start = timezone.now() - timedelta(seconds=monitor_settings.LOGIN_ATTEMPT_WINDOW)
        failures = LoginAttempt.objects.filter(
            ip_address=ip, success=False, timestamp__gte=window_start
        ).count()

    if monitor_settings.MAX_LOGIN_ATTEMPTS is not None and failures >= monitor_settings.MAX_LOGIN_ATTEMPTS:
        SecurityEvent.objects.create(
            ip_address=ip,
            event_type='brute_force',
            severity='critical',
            path=request.path,
            method=request.method,
            user_agent=ua,
            threat_score_delta=w.get('brute_force', 25),
            extra_data={
                'username': username,
                'failures_in_window': failures,
                'superuser_test': is_superuser,  # ✅ tag it
            },
        )
        _bump_score(ip, w.get('brute_force', 25), is_superuser=is_superuser)
    else:
        SecurityEvent.objects.create(
            ip_address=ip,
            event_type='auth_failure',
            severity='medium',
            path=request.path,
            method=request.method,
            user_agent=ua,
            threat_score_delta=w.get('auth_failure', 5),
            extra_data={
                'username': username,
                'superuser_test': is_superuser,  # ✅ tag it
            },
        )
        _bump_score(ip, w.get('auth_failure', 5), is_superuser=is_superuser)



@receiver(user_logged_in)
def on_login_success(sender, user, request, **kwargs):
    if not request:
        return
    ip = _get_ip(request)
    ua = request.META.get('HTTP_USER_AGENT', '')[:500]
    LoginAttempt.objects.create(
        ip_address=ip, username=user.username, success=True, user_agent=ua
    )


def _bump_score(ip, delta, is_superuser=False):
    if not ip or delta == 0:
        return
    threat, created = ThreatScore.objects.get_or_create(
        ip_address=ip, defaults={'score': delta, 'event_count': 1}
    )
    if not created:
        ThreatScore.objects.filter(pk=threat.pk).update(
            score=F('score') + delta,
            event_count=F('event_count') + 1,
        )
    threat.refresh_from_db()

    # ✅ Never auto-block a superuser
    if is_superuser:
        return

    if (
        monitor_settings.AUTO_BLOCK
        and monitor_settings.BLOCK_THRESHOLD is not None
        and threat.score >= monitor_settings.BLOCK_THRESHOLD
        and not threat.is_blocked
    ):
        from django.utils import timezone
        ThreatScore.objects.filter(pk=threat.pk).update(
            is_blocked=True,
            blocked_at=timezone.now(),
            block_reason=f'Auto-blocked: score ≥ {monitor_settings.BLOCK_THRESHOLD}',
        )





from functools import wraps
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect

from .conf import monitor_settings


def security_monitor_required(view_func):
    """
    Grants access if:
      - SECURITY_MONITOR['ALLOWED_USERS'] is set → user.username must be in that list
      - Otherwise → user must be a superuser
    """
    @wraps(view_func)
    def wrapper(request, *args, **kwargs):
        if not request.user.is_authenticated:
            login_url = getattr(settings, 'LOGIN_URL', '/accounts/login/')
            return redirect(f"{login_url}?next={request.path}")

        allowed_users = monitor_settings.ALLOWED_USERS
        if allowed_users:
            if request.user.username not in allowed_users:
                raise PermissionDenied
        else:
            if not request.user.is_superuser:
                raise PermissionDenied

        return view_func(request, *args, **kwargs)
    return wrapper



from django.urls import path
from . import views

app_name = 'security_monitor'

urlpatterns = [
    path('', views.dashboard, name='dashboard'),
    path('events/', views.events_list, name='events'),
    path('threats/', views.threats_list, name='threats'),
    path('threats/<str:ip_address>/', views.ip_detail, name='ip_detail'),
    path('visitors/', views.visitors_list, name='visitors'),

    # Actions
    path('block/<str:ip_address>/', views.block_ip, name='block_ip'),
    path('unblock/<str:ip_address>/', views.unblock_ip, name='unblock_ip'),
    path('whitelist/<str:ip_address>/', views.whitelist_ip, name='whitelist_ip'),
    path('whitelist/<str:ip_address>/remove/', views.remove_whitelist, name='remove_whitelist'),
    path('events/<int:event_id>/delete/', views.delete_event, name='delete_event'),

    # JSON APIs
    path('api/live-events/', views.live_events, name='live_events'),
    path('api/stats/', views.stats_api, name='stats_api'),
]



import json
from datetime import timedelta

from django.contrib import messages
from django.core.paginator import Paginator
from django.db.models import Count, Q
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils import timezone
from django.views.decorators.http import require_POST

from .conf import monitor_settings
from .decorators import security_monitor_required
from .models import (
    HoneypotHit, IPWhitelist, LoginAttempt,
    PageView, SecurityEvent, ThreatScore, Visitor,
)


# ── helpers ──────────────────────────────────────────────────────────────────
def _now():
    return timezone.now()

def _today_start():
    return _now().replace(hour=0, minute=0, second=0, microsecond=0)


# ── Dashboard ────────────────────────────────────────────────────────────────
@security_monitor_required
def dashboard(request):
    now = _now()
    today = _today_start()
    week_ago = now - timedelta(days=7)

    # Stat cards
    stats = {
        'events_today': SecurityEvent.objects.filter(timestamp__gte=today).count(),
        'critical_today': SecurityEvent.objects.filter(
            timestamp__gte=today, severity='critical'
        ).count(),
        'blocked_ips': ThreatScore.objects.filter(is_blocked=True).count(),
        'active_threats': ThreatScore.objects.filter(
            score__gte=20, is_blocked=False
        ).count(),
        'total_visitors': Visitor.objects.count(),
        'honeypot_hits': HoneypotHit.objects.filter(timestamp__gte=today).count(),
    }

    # Recent events feed (last 30)
    recent_events = SecurityEvent.objects.select_related('user').order_by('-timestamp')[:30]

    # Top threat IPs
    top_threats = ThreatScore.objects.order_by('-score')[:15]

    # Events by type (for donut chart)
    events_by_type = list(
        SecurityEvent.objects.filter(timestamp__gte=week_ago)
        .values('event_type')
        .annotate(count=Count('id'))
        .order_by('-count')
    )

    # Events per day for last 7 days (for line chart)
    events_per_day = []
    for i in range(6, -1, -1):
        day_start = (now - timedelta(days=i)).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        day_end = day_start + timedelta(days=1)
        events_per_day.append({
            'date': day_start.strftime('%b %d'),
            'count': SecurityEvent.objects.filter(
                timestamp__gte=day_start, timestamp__lt=day_end
            ).count(),
        })

    # Severity breakdown (last 24h)
    severity_counts = dict(
        SecurityEvent.objects.filter(timestamp__gte=now - timedelta(hours=24))
        .values('severity')
        .annotate(c=Count('id'))
        .values_list('severity', 'c')
    )

    # Top attacked paths
    top_paths = list(
        SecurityEvent.objects.filter(timestamp__gte=week_ago)
        .values('path')
        .annotate(count=Count('id'))
        .order_by('-count')[:10]
    )

    context = {
        'stats': stats,
        'recent_events': recent_events,
        'top_threats': top_threats,
        'events_by_type_json': json.dumps(events_by_type),
        'events_per_day_json': json.dumps(events_per_day),
        'severity_counts': severity_counts,
        'top_paths': top_paths,
        'page': 'dashboard',
    }
    return render(request, 'django_security_monitor/dashboard.html', context)


# ── Events list ──────────────────────────────────────────────────────────────
@security_monitor_required
def events_list(request):
    qs = SecurityEvent.objects.select_related('user').order_by('-timestamp')

    # Filters
    severity = request.GET.get('severity')
    event_type = request.GET.get('event_type')
    ip = request.GET.get('ip', '').strip()
    days = request.GET.get('days', '7')

    try:
        days = int(days)
    except ValueError:
        days = 7

    qs = qs.filter(timestamp__gte=_now() - timedelta(days=days))

    if severity:
        qs = qs.filter(severity=severity)
    if event_type:
        qs = qs.filter(event_type=event_type)
    if ip:
        qs = qs.filter(ip_address__icontains=ip)

    paginator = Paginator(qs, 50)
    page_obj = paginator.get_page(request.GET.get('page'))

    context = {
        'page_obj': page_obj,
        'event_types': SecurityEvent.EVENT_TYPES,
        'severities': SecurityEvent.SEVERITY,
        'period_options': [1, 7, 30, 90],
        'current_filters': {
            'severity': severity, 'event_type': event_type,
            'ip': ip, 'days': days,
        },
        'page': 'events',
    }
    return render(request, 'django_security_monitor/events.html', context)


# ── Threats list ─────────────────────────────────────────────────────────────
@security_monitor_required
def threats_list(request):
    qs = ThreatScore.objects.all()

    show = request.GET.get('show', 'all')
    if show == 'blocked':
        qs = qs.filter(is_blocked=True)
    elif show == 'active':
        qs = qs.filter(is_blocked=False, score__gte=10)

    paginator = Paginator(qs.order_by('-score'), 40)
    page_obj = paginator.get_page(request.GET.get('page'))

    context = {
        'page_obj': page_obj,
        'show': show,
        'page': 'threats',
    }
    return render(request, 'django_security_monitor/threats.html', context)


# ── IP detail ─────────────────────────────────────────────────────────────────
@security_monitor_required
def ip_detail(request, ip_address):
    threat, _ = ThreatScore.objects.get_or_create(ip_address=ip_address)
    events_qs = SecurityEvent.objects.filter(ip_address=ip_address)
    events = events_qs.order_by('-timestamp')[:100]
    login_attempts = LoginAttempt.objects.filter(ip_address=ip_address).order_by('-timestamp')[:20]
    honeypot_hits = HoneypotHit.objects.filter(ip_address=ip_address).order_by('-timestamp')[:10]
    visitor = Visitor.objects.filter(ip_address=ip_address).first()
    whitelisted = IPWhitelist.objects.filter(ip_address=ip_address).exists()

    # Event type breakdown for this IP
    event_breakdown = list(
        events_qs.values('event_type').annotate(count=Count('id')).order_by('-count')
    )

    context = {
        'threat': threat,
        'events': events,
        'login_attempts': login_attempts,
        'honeypot_hits': honeypot_hits,
        'visitor': visitor,
        'whitelisted': whitelisted,
        'event_breakdown_json': json.dumps(event_breakdown),
        'page': 'threats',
    }
    return render(request, 'django_security_monitor/ip_detail.html', context)


# ── Visitors ─────────────────────────────────────────────────────────────────
@security_monitor_required
def visitors_list(request):
    qs = Visitor.objects.select_related('user').order_by('-last_visit')

    ip = request.GET.get('ip', '').strip()
    if ip:
        qs = qs.filter(ip_address__icontains=ip)

    paginator = Paginator(qs, 50)
    page_obj = paginator.get_page(request.GET.get('page'))

    context = {
        'page_obj': page_obj,
        'page': 'visitors',
    }
    return render(request, 'django_security_monitor/visitors.html', context)


# ── Actions (POST) ────────────────────────────────────────────────────────────
@security_monitor_required
@require_POST
def block_ip(request, ip_address):
    reason = request.POST.get('reason', 'Manually blocked by admin')
    threat, _ = ThreatScore.objects.get_or_create(ip_address=ip_address)
    ThreatScore.objects.filter(pk=threat.pk).update(
        is_blocked=True,
        blocked_at=timezone.now(),
        block_reason=reason,
    )
    messages.success(request, f"✅ {ip_address} has been blocked.")
    return redirect(request.META.get('HTTP_REFERER', 'security_monitor:threats'))


@security_monitor_required
@require_POST
def unblock_ip(request, ip_address):
    ThreatScore.objects.filter(ip_address=ip_address).update(
        is_blocked=False,
        block_reason='',
        blocked_at=None,
    )
    messages.success(request, f"✅ {ip_address} has been unblocked.")
    return redirect(request.META.get('HTTP_REFERER', 'security_monitor:threats'))


@security_monitor_required
@require_POST
def whitelist_ip(request, ip_address):
    reason = request.POST.get('reason', '')
    IPWhitelist.objects.get_or_create(
        ip_address=ip_address,
        defaults={'added_by': request.user, 'reason': reason},
    )
    # Also unblock if blocked
    ThreatScore.objects.filter(ip_address=ip_address).update(
        is_blocked=False, block_reason=''
    )
    messages.success(request, f"✅ {ip_address} has been whitelisted.")
    return redirect(request.META.get('HTTP_REFERER', 'security_monitor:threats'))


@security_monitor_required
@require_POST
def remove_whitelist(request, ip_address):
    IPWhitelist.objects.filter(ip_address=ip_address).delete()
    messages.success(request, f"🗑 {ip_address} removed from whitelist.")
    return redirect(request.META.get('HTTP_REFERER', 'security_monitor:threats'))


@security_monitor_required
@require_POST
def delete_event(request, event_id):
    SecurityEvent.objects.filter(pk=event_id).delete()
    return JsonResponse({'status': 'ok'})


# ── JSON APIs (for live polling) ──────────────────────────────────────────────
@security_monitor_required
def live_events(request):
    since_id = request.GET.get('since', 0)
    events = SecurityEvent.objects.filter(pk__gt=since_id).order_by('-timestamp')[:20]
    data = [
        {
            'id': e.pk,
            'ip': e.ip_address,
            'type': e.get_event_type_display(),
            'severity': e.severity,
            'path': e.path,
            'timestamp': e.timestamp.strftime('%H:%M:%S'),
        }
        for e in events
    ]
    return JsonResponse({'events': data})


@security_monitor_required
def stats_api(request):
    now = _now()
    today = _today_start()
    return JsonResponse({
        'events_today': SecurityEvent.objects.filter(timestamp__gte=today).count(),
        'critical_today': SecurityEvent.objects.filter(
            timestamp__gte=today, severity='critical'
        ).count(),
        'blocked_ips': ThreatScore.objects.filter(is_blocked=True).count(),
        'active_threats': ThreatScore.objects.filter(
            score__gte=20, is_blocked=False
        ).count(),
    })



from django.contrib import admin
from .models import (
    SecurityEvent, ThreatScore, LoginAttempt,
    IPWhitelist, HoneypotHit, Visitor, PageView,
)


@admin.register(SecurityEvent)
class SecurityEventAdmin(admin.ModelAdmin):
    list_display = ('timestamp', 'ip_address', 'event_type', 'severity', 'path', 'threat_score_delta')
    list_filter = ('severity', 'event_type')
    search_fields = ('ip_address', 'path', 'user_agent')
    readonly_fields = ('timestamp',)
    ordering = ('-timestamp',)


@admin.register(ThreatScore)
class ThreatScoreAdmin(admin.ModelAdmin):
    list_display = ('ip_address', 'score', 'risk_level', 'is_blocked', 'event_count', 'last_event')
    list_filter = ('is_blocked',)
    search_fields = ('ip_address',)
    readonly_fields = ('first_seen', 'last_event')
    actions = ['block_selected', 'unblock_selected']

    @admin.action(description='Block selected IPs')
    def block_selected(self, request, queryset):
        from django.utils import timezone
        queryset.update(
            is_blocked=True,
            blocked_at=timezone.now(),
            block_reason='Manually blocked via admin',
        )

    @admin.action(description='Unblock selected IPs')
    def unblock_selected(self, request, queryset):
        queryset.update(is_blocked=False, block_reason='', blocked_at=None)


@admin.register(IPWhitelist)
class IPWhitelistAdmin(admin.ModelAdmin):
    list_display = ('ip_address', 'added_by', 'reason', 'created_at')
    search_fields = ('ip_address',)


@admin.register(LoginAttempt)
class LoginAttemptAdmin(admin.ModelAdmin):
    list_display = ('timestamp', 'ip_address', 'username', 'success', 'user_agent')
    list_filter = ('success',)
    search_fields = ('ip_address', 'username')


@admin.register(HoneypotHit)
class HoneypotHitAdmin(admin.ModelAdmin):
    list_display = ('timestamp', 'ip_address', 'path', 'user_agent')
    search_fields = ('ip_address', 'path')


@admin.register(Visitor)
class VisitorAdmin(admin.ModelAdmin):
    list_display = ('ip_address', 'user', 'visit_count', 'first_visit', 'last_visit')
    search_fields = ('ip_address',)
    readonly_fields = ('first_visit', 'last_visit')



"""
Celery tasks — only registered if Celery is available.
Falls back to nothing if project has no Celery.
"""
from .conf import is_celery_available

if is_celery_available():
    from celery import shared_task
    from datetime import timedelta
    from django.utils import timezone
    from django.core.mail import send_mail
    from django.conf import settings
    from .models import SecurityEvent, ThreatScore
    from .conf import monitor_settings

    @shared_task(name='security_monitor.cleanup_old_events')
    def cleanup_old_events():
        """Delete events older than EVENT_RETENTION_DAYS."""
        cutoff = timezone.now() - timedelta(days=monitor_settings.EVENT_RETENTION_DAYS)
        deleted, _ = SecurityEvent.objects.filter(timestamp__lt=cutoff).delete()
        return f"Deleted {deleted} old security events"

    @shared_task(name='security_monitor.send_critical_alert')
    def send_critical_alert(event_id):
        """Email the admin when a critical event occurs."""
        alert_email = monitor_settings.ALERT_EMAIL
        if not alert_email:
            return "No ALERT_EMAIL configured"
        try:
            event = SecurityEvent.objects.get(pk=event_id)
            send_mail(
                subject=f"[SIEM CRITICAL] {event.get_event_type_display()} from {event.ip_address}",
                message=(
                    f"IP: {event.ip_address}\n"
                    f"Path: {event.path}\n"
                    f"Type: {event.get_event_type_display()}\n"
                    f"Time: {event.timestamp}\n"
                    f"Payload: {event.payload or 'N/A'}"
                ),
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[alert_email],
                fail_silently=True,
            )
            return f"Alert sent for event {event_id}"
        except SecurityEvent.DoesNotExist:
            return "Event not found"

    @shared_task(name='security_monitor.decay_threat_scores')
    def decay_threat_scores():
        """
        Gradually reduce threat scores over time so old IPs
        aren't permanently flagged.
        """
        from django.db.models import F
        cutoff = timezone.now() - timedelta(hours=24)
        updated = ThreatScore.objects.filter(
            last_event__lt=cutoff, is_blocked=False, score__gt=0
        ).update(score=F('score') * 0.9)
        return f"Decayed scores for {updated} IPs"
    # tasks.py
    @shared_task(name='security_monitor.log_pageview_async')
    def log_pageview_async(visitor_id, path, method, status_code, response_time_ms):
        from .models import PageView
        PageView.objects.create(
            visitor_id=visitor_id, path=path, method=method,
            status_code=status_code, response_time_ms=response_time_ms,
        )


from django.apps import AppConfig


class DjangoSecurityMonitorConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_security_monitor'
    verbose_name = 'Security Monitor'


    
    def ready(self):
        import django_security_monitor.signals  # noqa: F401 — register signal handlers


import math
from django.db import models
from django.conf import settings
from django.utils import timezone


class Visitor(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL, null=True, blank=True
    )
    session_key = models.CharField(max_length=40, unique=True)
    ip_address = models.CharField(max_length=45)
    user_agent = models.CharField(max_length=500, blank=True, null=True)
    first_visit = models.DateTimeField(auto_now_add=True)
    last_visit = models.DateTimeField(auto_now=True)
    visit_count = models.PositiveIntegerField(default=1)
    location = models.JSONField(blank=True, null=True)

    class Meta:
        indexes = [
            models.Index(fields=['ip_address']),
            models.Index(fields=['last_visit']),
        ]

    def __str__(self):
        return f"{self.ip_address} — {self.visit_count} visits"


class PageView(models.Model):
    visitor = models.ForeignKey(Visitor, on_delete=models.CASCADE, related_name='page_views')
    path = models.CharField(max_length=500)
    method = models.CharField(max_length=10, default='GET')
    status_code = models.PositiveSmallIntegerField(null=True, blank=True)
    response_time_ms = models.PositiveIntegerField(null=True, blank=True)
    timestamp = models.DateTimeField(auto_now_add=True)

    class Meta:
        indexes = [
            models.Index(fields=['timestamp']),
            models.Index(fields=['path']),
        ]

    def __str__(self):
        return f"{self.visitor.ip_address} → {self.path}"


class SecurityEvent(models.Model):
    EVENT_TYPES = [
        ('suspicious_path', 'Suspicious Path'),
        ('sensitive_file', 'Sensitive File Access'),
        ('brute_force', 'Brute Force Attack'),
        ('rate_limit', 'Rate Limit Exceeded'),
        ('sql_injection', 'SQL Injection'),
        ('xss_attempt', 'XSS Attempt'),
        ('path_traversal', 'Path Traversal'),
        ('scanner', 'Security Scanner'),
        ('auth_failure', 'Auth Failure'),
        ('csrf_failure', 'CSRF Failure'),
        ('honeypot', 'Honeypot Triggered'),
        ('403', '403 Forbidden'),
        ('404', '404 Not Found'),
        ('500', '500 Server Error'),
        ('blocked_ip', 'Blocked IP Attempt'),
    ]
    SEVERITY = [
        ('info', 'Info'),
        ('low', 'Low'),
        ('medium', 'Medium'),
        ('high', 'High'),
        ('critical', 'Critical'),
    ]

    ip_address = models.CharField(max_length=45, db_index=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True,
        on_delete=models.SET_NULL, related_name='security_events'
    )
    event_type = models.CharField(max_length=30, choices=EVENT_TYPES, db_index=True)
    severity = models.CharField(max_length=10, choices=SEVERITY, default='medium', db_index=True)
    path = models.CharField(max_length=500)
    method = models.CharField(max_length=10, default='GET')
    user_agent = models.CharField(max_length=500, blank=True)
    referer = models.CharField(max_length=500, blank=True)
    payload = models.TextField(blank=True)
    threat_score_delta = models.FloatField(default=0)
    extra_data = models.JSONField(default=dict, blank=True)
    timestamp = models.DateTimeField(auto_now_add=True, db_index=True)

    class Meta:
        ordering = ['-timestamp']
        indexes = [
            models.Index(fields=['ip_address', 'timestamp']),
            models.Index(fields=['event_type', 'timestamp']),
            models.Index(fields=['severity', 'timestamp']),
        ]

    def __str__(self):
        return f"[{self.severity.upper()}] {self.get_event_type_display()} — {self.ip_address}"


class ThreatScore(models.Model):
    ip_address = models.CharField(max_length=45, unique=True, db_index=True)
    score = models.FloatField(default=0, db_index=True)
    event_count = models.PositiveIntegerField(default=0)
    first_seen = models.DateTimeField(auto_now_add=True)
    last_event = models.DateTimeField(auto_now=True)
    is_blocked = models.BooleanField(default=False, db_index=True)
    block_reason = models.CharField(max_length=255, blank=True)
    blocked_at = models.DateTimeField(null=True, blank=True)
    location = models.JSONField(blank=True, null=True)
    notes = models.TextField(blank=True)

    class Meta:
        ordering = ['-score']

    @property
    def risk_level(self):
        if self.score >= 80:
            return 'critical'
        if self.score >= 50:
            return 'high'
        if self.score >= 20:
            return 'medium'
        return 'low'

    @property
    def hack_probability(self):
        """Sigmoid curve: 0–100 probability this IP is malicious."""
        try:
            p = 100 / (1 + math.exp(-0.08 * (self.score - 40)))
            return round(min(100.0, max(0.0, p)), 1)
        except OverflowError:
            return 100.0

    def __str__(self):
        tag = 'BLOCKED' if self.is_blocked else self.risk_level.upper()
        return f"{self.ip_address} — Score: {self.score:.1f} [{tag}]"


class LoginAttempt(models.Model):
    ip_address = models.CharField(max_length=45, db_index=True)
    username = models.CharField(max_length=150)
    success = models.BooleanField(default=False)
    user_agent = models.CharField(max_length=500, blank=True)
    timestamp = models.DateTimeField(auto_now_add=True, db_index=True)

    class Meta:
        indexes = [
            models.Index(fields=['ip_address', 'timestamp']),
            models.Index(fields=['username', 'timestamp']),
        ]

    def __str__(self):
        return f"{'OK' if self.success else 'FAIL'}: {self.username} @ {self.ip_address}"


class IPWhitelist(models.Model):
    ip_address = models.CharField(max_length=45, unique=True)
    added_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True
    )
    reason = models.CharField(max_length=255, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"Whitelisted: {self.ip_address}"


class HoneypotHit(models.Model):
    """Accesses to trap URLs — only scanners/bots hit these."""
    ip_address = models.CharField(max_length=45, db_index=True)
    path = models.CharField(max_length=500)
    user_agent = models.CharField(max_length=500, blank=True)
    headers = models.JSONField(default=dict, blank=True)
    timestamp = models.DateTimeField(auto_now_add=True, db_index=True)

    class Meta:
        ordering = ['-timestamp']

    def __str__(self):
        return f"Honeypot: {self.ip_address} → {self.path}"



from django.conf import settings as django_settings

DEFAULTS = {
    'ALLOWED_USERS': [],            # empty = superusers only; or ['alice', 'bob']
    'SUSPICIOUS_PATHS': [
        '.env', '.git', '.svn', 'wp-admin', 'wp-login',
        'phpMyAdmin', 'phpmyadmin', '/etc/passwd', '/etc/shadow',
        'config.php', '.htaccess', '.htpasswd', 'web.config',
        '.DS_Store', 'xmlrpc.php', '/proc/', 'backup.sql', 'dump.sql',
    ],
    'SENSITIVE_EXTENSIONS': [
        '.env', '.sql', '.bak', '.backup', '.dump', '.config',
        '.conf', '.key', '.pem', '.p12', '.pfx', '.log',
        '.db', '.sqlite', '.sqlite3',
    ],
    'HONEYPOT_PATHS': [
        '/wp-login.php', '/wp-admin/', '/admin/login.php',
        '/.env', '/.git/config', '/config.php',
        '/phpinfo.php', '/server-status', '/admin.php',
        '/shell.php', '/c99.php', '/r57.php',
    ],
    'BLOCK_THRESHOLD': None,
    'RATE_LIMIT': None,
    'RATE_LIMIT_WINDOW': None,
    'MAX_LOGIN_ATTEMPTS': None,
    'LOGIN_ATTEMPT_WINDOW': None,
    'USE_REDIS': None,              # None = auto-detect
    'USE_CELERY': None,             # None = auto-detect
    'DASHBOARD_URL': 'security-monitor/',
    'ALERT_EMAIL': None,
    'LOG_404': True,
    'LOG_500': True,
    'EXCLUDE_PATHS': ['/static/', '/media/'],
    'WHITELIST_CACHE_TTL': 300,
    'EVENT_RETENTION_DAYS': 90,
    'THREAT_SCORE_WEIGHTS': {
        'suspicious_path': 15,
        'sensitive_file': 20,
        'rate_limit': 10,
        'auth_failure': 5,
        'brute_force': 25,
        'sql_injection': 30,
        'xss_attempt': 25,
        'path_traversal': 35,
        'scanner': 20,
        'honeypot': 40,
        '404': 3,
        '500': 2,
        'csrf_failure': 15,
        'blocked_ip': 0,
    },
    'AUTO_BLOCK': False,
}


class _SecurityMonitorSettings:
    def __init__(self):
        self._cache = None

    def _load(self):
        self._cache = getattr(django_settings, 'SECURITY_MONITOR', {})

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(name)
        if self._cache is None:
            self._load()
        if name in self._cache:
            return self._cache[name]
        if name in DEFAULTS:
            return DEFAULTS[name]
        raise AttributeError(f"Invalid SECURITY_MONITOR setting: '{name}'")

    def get(self, name, default=None):
        try:
            return getattr(self, name)
        except AttributeError:
            return default


monitor_settings = _SecurityMonitorSettings()


def is_redis_available():
    if monitor_settings.USE_REDIS is False:
        return False
    try:
        backend = django_settings.CACHES.get('default', {}).get('BACKEND', '')
        if 'redis' not in backend.lower():
            return False
        from django.core.cache import cache
        cache.set('_sm_probe', 1, 1)
        return cache.get('_sm_probe') == 1
    except Exception:
        return False


def is_celery_available():
    if monitor_settings.USE_CELERY is False:
        return False
    try:
        from celery import current_app
        return current_app.configured
    except Exception:
        return False
